<?php
namespace nur\sery\db\sqlite;

use nur\sery\cl;
use nur\sery\db\AbstractCapacitor;
use nur\sery\db\CapacitorChannel;
use nur\sery\php\func;
use nur\sery\ValueException;

/**
 * Class SqliteCapacitor
 */
class SqliteCapacitor extends AbstractCapacitor{
  function __construct($sqlite) {
    $this->sqlite = Sqlite::with($sqlite);
  }

  /** @var Sqlite */
  protected $sqlite;

  function sqlite(): Sqlite {
    return $this->sqlite;
  }

  protected function _create(CapacitorChannel $channel): void {
    if (!$channel->isCreated()) {
      $columns = cl::merge([
        "_id" => "integer primary key autoincrement",
        "_item" => "text",
      ], $channel->getKeyDefinitions());
      $this->sqlite->exec([
        "create table if not exists",
        "table" => $channel->getTableName(),
        "cols" => $columns,
      ]);
      $channel->setCreated();
    }
  }

  /** @var CapacitorChannel[] */
  protected $channels;

  function addChannel(CapacitorChannel $channel): CapacitorChannel {
    $this->_create($channel);
    $this->channels[$channel->getName()] = $channel;
    return $channel;
  }

  protected function getChannel(?string $name=null): CapacitorChannel {
    $name = CapacitorChannel::verifix_name($name);
    $channel = $this->channels[$name] ?? null;
    if ($channel === null) {
      $channel = $this->addChannel(new CapacitorChannel($name));
    }
    return $channel;
  }

  function _exists(CapacitorChannel $channel): bool {
    $tableName = $this->sqlite->get([
      "select name from sqlite_schema",
      "where" => [
        "name" => $channel->getTableName(),
      ],
    ]);
    return $tableName !== null;
  }

  function _reset(CapacitorChannel $channel): void {
    $this->sqlite->exec([
      "drop table if exists",
      $channel->getTableName(),
    ]);
    $channel->setCreated(false);
  }

  function _charge($item, CapacitorChannel $channel): void {
    $this->_create($channel);
    $values = cl::merge($channel->getKeyValues($item), [
      "_item" => serialize($item),
    ]);
    $this->sqlite->exec([
      "insert",
      "into" => $channel->getTableName(),
      "values" => $values,
    ]);
  }

  function _discharge($keys=null, CapacitorChannel $channel=null, ?bool $reset=null): iterable {
    if ($keys !== null && !is_array($keys)) $keys = ["_id" => $keys];
    if ($reset === null) $reset = $keys === null;
    $rows = $this->sqlite->all([
      "select _item",
      "from" => $channel->getTableName(),
      "where" => $keys,
    ]);
    foreach ($rows as $row) {
      $item = unserialize($row['_item']);
      yield $item;
    }
    if ($reset) $this->_reset($channel);
  }

  function _get($keys, CapacitorChannel $channel=null) {
    if ($keys === null) throw ValueException::null("keys");
    if (!is_array($keys)) $keys = ["_id" => $keys];
    $row = $this->sqlite->one([
      "select _item",
      "from" => $channel->getTableName(),
      "where" => $keys,
    ]);
    if ($row === null) return null;
    else return unserialize($row["_item"]);
  }

  function _each($keys, callable $func, ?array $args=null, CapacitorChannel $channel=null): void {
    if ($keys !== null && !is_array($keys)) $keys = ["_id" => $keys];
    $context = func::_prepare($func);
    $sqlite = $this->sqlite;
    $tableName = $channel->getTableName();
    $rows = $sqlite->all([
      "select",
      "from" => $tableName,
      "where" => $keys,
    ]);
    $args ??= [];
    foreach ($rows as $row) {
      $item = unserialize($row['_item']);
      $updates = func::_call($context, [$item, $row, ...$args]);
      if (is_array($updates)) {
        if (array_key_exists("_item", $updates)) {
          $updates["_item"] = serialize($updates["_item"]);
        }
        $sqlite->exec([
          "update",
          "table" => $tableName,
          "values" => $updates,
          "where" => ["_id" => $row["_id"]],
        ]);
      }
    }
  }

  function close(): void {
    $this->sqlite->close();
  }
}
